#include "Controller.h"
#include <string>
#include <sstream>
#include <iostream>
#include <list>

#include <RL_PP/State.h>
#include <RL_PP/AbstractAction.h>
#include <RL_PP/ActionTileCoding.h>
#include <RL_PP/ActionTilings.h>
#include <RL_PP/ActionLookupTable.h>
#include <RL_PP/Utility.h>


static const int MAXTORQUE = 200;
static const int RESOLUTION = 1; // "Auflösung" des Aktionsraumes
static const int ACTIONCOUNT = (2*RESOLUTION+1)*(2*RESOLUTION+1); // 2 Gelenkpaare
static const bool loadValueFunction = false;



/*
Die Klasse Action ist die einfachste Implementation einer Aktion für die Aufgabe.
Möglich wäre z.B. die torque für das Kommando erst aus einem besseren/abstrakten
 Parameterzu berechnen.
*/
class Action : public AbstractAction
{
  public:
    Action( int t ) : AbstractAction()
    {
      const int S = RESOLUTION*2+1;
      y = t/S;
      t -= y*S;
      z = t;
      y -= RESOLUTION;
      z -= RESOLUTION;
      y *= MAXTORQUE/RESOLUTION;
      z *= MAXTORQUE/RESOLUTION;
      std::cout << "used torques = " << y << " " << z << "\n";
    }

    inline const std::string command() const
    {
      std::ostringstream tmp;
      tmp << "(torque 14 " << y << ")\n";
      tmp << "(torque 15 " << -y << ")\n";
      tmp << "(torque 16 " << z << ")\n";
      tmp << "(torque 17 " << -z << ")\n";
      return tmp.str();
    }

  private:
    int y,z;
};



int Controller::theSocket;


Controller::Controller( const int aSocket)
  : Watkins( "rl.conf" )
{
  theSocket = aSocket;
  episodes = 0;

  if ( loadValueFunction ) {
    load();
    // hier die Aktionen setzen
    for( int j=0; j< ACTIONCOUNT; ++j ) {
      avfPtr->setAction( j, new Action( j ) );
    }
  }
  else {
    State* statePtr = new State();
    // Zustandsvariablen mit MIN- und MAX-Wert registrieren
    pos2Id = statePtr->registerVariable( "ANGLE2", -6.0, 10.0 );
    pos3Id = statePtr->registerVariable( "ANGLE3", -6.0, 10.0 );
  
    // ValueFunction anlegen
    // hier "ActionTileCoding"
    avfPtr = new ActionTileCoding( statePtr );
    // hier die Aktionen bei der valueFunction registrieren
    for( int j=0; j<ACTIONCOUNT; ++j ) {
      avfPtr->registerAction( new Action( j ) );
    }
    // Zusandsraum festlegen
    // hier mit TileCodings
    Types::TILINGS_VAR_STRUCT vs;
    std::list< Types::TILINGS_VAR_STRUCT > vsList;

    vsList.clear();
    vs.id = pos2Id;
    vs.resolution = 2.0; // gewünschte Auflösung
    vs.tileType = Types::CLIP;
    vsList.push_back( vs );
    ActionTilings* t = new ActionTilings( 2, vsList ); //Anzahl der Tiles
    static_cast< ActionTileCoding* >(avfPtr)->registerTiling( t );

    vsList.clear();
    vs.id = pos3Id;
    vs.resolution = 2.0; // gewünschte Auflösung
    vs.tileType = Types::CLIP;
    vsList.push_back( vs );
    t = new ActionTilings( 2, vsList ); //Anzahl der Tiles
    static_cast< ActionTileCoding* >(avfPtr)->registerTiling( t );
  }
}


void Controller::firstInit()
{
  std::cout << "++++++++ FIRST_INIT +++++++++++\n";
  recv ( theSocket, buffer, 1024, 0 );
  //alle joints in Ausgangslage fahren und fixieren
  sendMsg("(reset)\n");
  sendMsg("(pidmaxtorque** 800)\n");
  sendMsg("(pid* 512 202 800 202 800 512 512 512 512 512 512 512 512 512 512 512 512 512 512)\n");
  for ( int i = 0; i <= 200; ++i ) {
    sendMsg("(done)\n");
    recv ( theSocket, buffer, 1024, 0 );
  }
  sendMsg("(save)\n");
  sendMsg("(done)\n");
}




void Controller::init()
{
  recv ( theSocket, buffer, 1024, 0 );
  sendMsg("(reset)\n");
  sendMsg("(done)\n");
  recv ( theSocket, buffer, 1024, 0 );
  sendMsg("(restore)\n");
  sendMsg("(done)\n");
  recv ( theSocket, buffer, 1024, 0 );
  sendMsg("(pidmaxtorque** 800)\n");
  sendMsg("(pid* 512 202 800 202 800 512 512 512 512 512 512 512 512 512 512 512 512 512 512)\n");
  sendMsg("(done)\n");

  sprintf(buffer, "");
  len = recv ( theSocket, buffer, 1024, 0 );
  buffer[len] = '\0';

  // initiale Hüftposition parsen
  sscanf( strchr(buffer, (int)'p'), "pos hip  %f %f %f %*s", &initX, &initY, &initZ );

  // Beine weich machen
  sendMsg("(pid* 512 202 800 202 800 512 512 512 512 512 512 512 512 -1 -1 -1 -1 512 512)\n");
  sendMsg("(done)\n");
}


void Controller::run()
{
  // Aufruf der RL++-Kontrollschleife
  firstInit();
  controlLoop();
}



void Controller::setInitialState()
{
  init();
  steps=0;
  ++episodes;
  int a,b,c,d,e,f; // die 6 involvierten Gelenkwinkel
  //Umweltinformationen empfangen
  sprintf(buffer, "");
  int len = recv ( theSocket, buffer, 1024, 0 );
  buffer[len] = '\0';
  // Gelenkwinkel parsen
  sscanf( buffer, "(time %*f)(angle %*i %*i %*i %*i %*i %*i %*i %*i %*i %*i %*i %i %i %i %i %i %i %*s", &a, &b, &c, &d, &e, &f );
  // Hüftposition parsen
  sscanf( strchr(buffer, (int)'p'), "pos hip  %f %f %f %*s", &theX, &theY, &theZ );

  // neue Werte der Zustandsvariablen setzen
  avfPtr->getStatePtr()->setValueDbl( pos2Id, double(512-c) );
  avfPtr->getStatePtr()->setValueDbl( pos3Id, double(512-e) );
}


void Controller::takeAction()
{
  const Action* action = static_cast< const Action* >( avfPtr->getActiveAction() );
  if ( action == NULL ) {
    std::cerr << "USERFEHLER: avfPtr->getActiveAction() liefert NULL !!!\n";
    exit(-1);
  }
  sendMsg(action->command().c_str());
  sendMsg("(done)\n");
}


const bool Controller::observeState()
{
  ++steps;
  int a,b,c,d,e,f; // die 6 involvierten Gelenkwinkel

  //Umweltinformationen empfangen
  sprintf(buffer, "");
  int len = recv ( theSocket, buffer, 1024, 0 );
  buffer[len] = '\0';
  // Gelenkwinkel parsen
  sscanf( buffer, "(time %*f)(angle %*i %*i %*i %*i %*i %*i %*i %*i %*i %*i %*i %i %i %i %i %i %i %*s", &a, &b, &c, &d, &e, &f );
  // Hüftposition parsen
  sscanf( strchr(buffer, (int)'p'), "pos hip  %f %f %f %*s", &theX, &theY, &theZ );

  // neue Werte der Zustandsvariablen setzen
  avfPtr->getStatePtr()->setValueDbl( pos2Id, double(512-c) );
  avfPtr->getStatePtr()->setValueDbl( pos3Id, double(512-e) );

  float hipDiff = sqrt( (initX-theX)*(initX-theX) + (initY-theY)*(initY-theY) + (initZ-theZ)*(initZ-theZ));

  if ( (steps > 2000) || ( hipDiff > 0.04) ) { //max. 20 Sekunden
    sendMsg("(done)\n");
    return false;
  }
  return true;
}


const double Controller::calculateReward()
{
  double reward = 0.0;
  // Differenz zur Zielposition bestimmen
  float diff = sqrt( (initX-theX)*(initX-theX) + (initY-theY)*(initY-theY) + (initZ-theZ)*(initZ-theZ));
  // Wenn Zielposition erreicht, gibt's Belohnung
  if (diff < 0.005) {
    reward = 1;
  }
  // wenn deutlich verfehlt, wird bestraft
  if (diff > 0.02) {
    reward = -10;
  }
  // sonst nicht
  return reward;
}


void Controller::finishEpisode()
{
  std::cout << "episodes= " << episodes << " steps = " << steps << "\n";
  // valueFunction speichern
  if ( (episodes%10) == 0 ) {
    save();
    ALPHA *= 0.95;
    std::cout << "ALPHA = " << ALPHA << "\n";
  }
}


void Controller::save() const
{
  // save valuefunction
  std::cerr << "saving ... ";
  int res = avfPtr->save( "aufgabe5.vf" );
  if (res==0) {
    std::cerr << "okay\n";
  }
  else {
    std::cerr << "failed\n";
  }
}

void Controller::load()
{
  // load valuefunction
  avfPtr = new ActionTileCoding( "aufgabe5.vf" );
  avfPtr->setEpisodeStateHistory( true );
  pos2Id = avfPtr->getStatePtr()->getVariableId( "ANGLE2" );
  pos3Id = avfPtr->getStatePtr()->getVariableId( "ANGLE3" );
}
